/*-
 * Copyright 2015-2016 UPLEX - Nils Goroll Systemoptimierung
 * All rights reserved.
 *
 * Authors: Nils Goroll <nils.goroll@uplex.de>
 *          Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <errno.h>
#include <stdint.h>

#include "base64.h"

#include "vrt.h"
#include "vas.h"

static inline int
decode(char *restrict *restrict dest, char *restrict const buf,
       const size_t maxlen, unsigned u, const int n)
{
	char *d;

	if (n <= 1)
		return -1;
	d = *dest;
	for (int i = 0; i < n - 1; i++) {
		if (d == buf + maxlen)
			return -1;
		*d++ = (u >> 16) & 0xff;
		u <<= 8;
	}
	*dest += d - *dest;
	return 1;
}

ssize_t
base64_encode(const enum encoding enc, char *restrict const buf,
	      const size_t maxlen, const char *restrict const inbuf,
	      const size_t inlength)
{
	const struct b64_alphabet *alpha = &b64_alphabet[enc];
	char *p = buf;
	const uint8_t *in = (uint8_t *)inbuf;

	AN(buf);
	AN(alpha);
	if (in == NULL || inlength == 0)
		return 0;
	if ((enc != BASE64URLNOPAD && maxlen < base64_encode_l(inlength))
	    || maxlen < base64nopad_encode_l(inlength))
		return -1;

	while (in < (uint8_t *)inbuf + inlength) {
		uint8_t idx;
		size_t rest = ((uint8_t *)inbuf + inlength) - in;

		*p++ = alpha->b64[(in[0] >> 2) & 0x3f];

		idx = (in[0] << 4);
		if (rest > 1)
			idx += (in[1] >> 4);
		*p++ = alpha->b64[idx & 0x3f];

		if (rest > 1) {
			idx = (in[1] << 2);
			if (rest > 2)
				idx += in[2] >> 6;
			*p++ = alpha->b64[idx & 0x3f];
		}
		else if (alpha->padding)
			*p++ = alpha->padding;

		if (rest > 2)
			*p++ = alpha->b64[in[2] & 0x3f];
		else if (alpha->padding)
			*p++ = alpha->padding;

		in += 3;
	}

	assert(p >= buf && p - buf <= maxlen);
	return p - buf;
}

ssize_t
base64_decode(const enum encoding dec, char *restrict const buf,
	      const size_t maxlen, const char *restrict const p, va_list ap)
{
	const struct b64_alphabet *alpha = &b64_alphabet[dec];
	char *dest = buf;
	unsigned u = 0, term = 0;
	int n = 0;

	AN(buf);
	AN(alpha);

	for (const char *s = p; s != vrt_magic_string_end;
	     s = va_arg(ap, const char *)) {
		if (s == NULL)
			continue;
		if (*s && term) {
			errno = EINVAL;
			return -1;
		}
		while (*s) {
			while (n < 4) {
				char b = alpha->i64[(unsigned) *s++];
				u <<= 6;
				if (b == ILL) {
					errno = EINVAL;
					return -1;
				}
				n++;
				if (b == PAD) {
					term++;
					continue;
				}
				u |= (unsigned) b;
				if (!*s)
					break;
			}
			if (n == 4) {
				if (decode(&dest, buf, maxlen, u, n-term) < 0) {
					errno = ENOMEM;
					return -1;
				}
				n = 0;
			}
		}
	}
	if (n) {
		if (!alpha->padding)
			u <<= 6 * (4 - n);
		if (decode(&dest, buf, maxlen, u, n-term) < 0) {
			errno = ENOMEM;
			return -1;
		}
	}

	return dest - buf;
}
