/*-
 * Copyright 2015-2016 UPLEX - Nils Goroll Systemoptimierung
 * All rights reserved.
 *
 * Authors: Nils Goroll <nils.goroll@uplex.de>
 *          Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <errno.h>
#include <pthread.h>

#include "config.h"

#include "vcl.h"
#include "vrt.h"

#include "vcc_if.h"
#include "vmod_blobcode.h"
#include "wb.h"

struct vmod_blobcode_blob {
	unsigned magic;
#define VMOD_BLOBCODE_BLOB_MAGIC 0xfade4fa9
	struct vmod_priv blob;
	const char *encoding[__MAX_ENCODING];
	pthread_mutex_t lock;
};

#define ERR(ctx, msg) \
	errmsg((ctx), "vmod blobcode error: " msg)

#define VERR(ctx, fmt, ...) \
	errmsg((ctx), "vmod blobcode error: " fmt, __VA_ARGS__)

#define ERRINVAL(ctx, enc) \
	VERR((ctx), "cannot decode, illegal encoding beginning with \"%s\"", \
	     (enc))

#define VERRNOMEM(ctx, fmt, ...) \
	VERR((ctx), fmt ", out of space", __VA_ARGS__)

#define ERRNOMEM(ctx, msg) \
	ERR((ctx), msg ", out of space")

#define INIT_FINI(ctx) (((ctx)->method & (VCL_MET_INIT | VCL_MET_FINI)) != 0)

#define ILLEGAL(ctx, m) \
	ERR((ctx), m " is illegal in vcl_init() and vcl_fini().")

static const struct vmod_priv const null_blob[1] =
{
	{
		.priv = "",
		.len = 0,
		.free = NULL
	}
};

static inline size_t
decode_l_va(enum encoding dec, const char * const p, va_list ap)
{
	size_t len = 0;
	const char *next;

	if (p == vrt_magic_string_end)
		return 0;
	if (p != NULL)
		len = strlen(p);

	while ((next = va_arg(ap, const char *)) != vrt_magic_string_end)
		if (next != NULL)
			len += strlen(next);

	switch(dec) {
	case IDENTITY:
		return len;
	case BASE64:
	case BASE64URL:
	case BASE64URLNOPAD:
		return base64_decode_l(len);
	case HEX:
		return hex_decode_l(len);
	default:
		WRONG("Illegal decoding");
	}
}

static inline ssize_t __match_proto__(decode_f)
decode(const enum encoding dec, char *restrict const buf,
       const size_t maxlen, const char *restrict const p, va_list ap)
{
	errno = 0;
	switch(dec) {
	case IDENTITY:
		return id_decode(dec, buf, maxlen, p, ap);
	case BASE64:
	case BASE64URL:
	case BASE64URLNOPAD:
		return base64_decode(dec, buf, maxlen, p, ap);
	case HEX:
		return hex_decode(dec, buf, maxlen, p, ap);
	default:
		WRONG("Illegal decoding");
	}
}

static inline ssize_t __match_proto__(encode_f)
encode(const enum encoding enc, char *restrict const buf, const size_t maxlen,
       const char *restrict const in, const size_t inlen)
{
	switch(enc) {
	case IDENTITY:
		return id_encode(enc, buf, maxlen, in, inlen);
	case BASE64:
	case BASE64URL:
	case BASE64URLNOPAD:
		return base64_encode(enc, buf, maxlen, in, inlen);
	case HEXUC:
	case HEXLC:
		return hex_encode(enc, buf, maxlen, in, inlen);
	default:
		WRONG("Illegal encoding");
	}
}

static void
errmsg(VRT_CTX, const char *fmt, ...)
{
	va_list args;

	va_start(args, fmt);
	if (ctx->method == VCL_MET_INIT) {
		AN(ctx->msg);
		VSB_vprintf(ctx->msg, fmt, args);
		VRT_handling(ctx, VCL_RET_FAIL);
	}
	else if (ctx->vsl)
		VSLbv(ctx->vsl, SLT_VCL_Error, fmt, args);
	else
		/* An error message in vcl_fini() cannot be formatted */
		VSL(SLT_VCL_Error, 0, fmt);
	va_end(args);
}

static void
err_decode(VRT_CTX, const char *enc)
{
	switch(errno) {
	case EINVAL:
		ERRINVAL(ctx, enc);
		break;
	case ENOMEM:
		ERRNOMEM(ctx, "cannot decode");
		break;
	default:
		WRONG("invalid errno");
	}
}

/* Objects */

VCL_VOID __match_proto__(td_blobcode_blob__init)
vmod_blob__init(VRT_CTX, struct vmod_blobcode_blob **blobp, const char *vcl_name,
		VCL_ENUM decs, const char *p, ...)
{
	struct vmod_blobcode_blob *b;
	enum encoding dec = parse_encoding(decs);
	va_list ap;
	ssize_t len;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	AN(blobp);
	AZ(*blobp);
	AN(vcl_name);
	ALLOC_OBJ(b, VMOD_BLOBCODE_BLOB_MAGIC);
	AN(b);
	*blobp = b;
	b->blob.free = NULL;
	AZ(pthread_mutex_init(&b->lock, NULL));

	va_start(ap, p);
	len = decode_l_va(dec, p, ap);
	va_end(ap);
	if (len == 0) {
		b->blob.len = 0;
		b->blob.priv = NULL;
		return;
	}
	assert(len > 0);

	b->blob.priv = malloc(len);
	if (b->blob.priv == NULL) {
		VERRNOMEM(ctx, "cannot create blob %s", vcl_name);
		return;
	}

	va_start(ap, p);
	len = decode(dec, b->blob.priv, len, p, ap);
	va_end(ap);

	if (len == -1) {
		assert(errno == EINVAL);
		free(b->blob.priv);
		b->blob.priv = NULL;
		AZ(pthread_mutex_destroy(&b->lock));
		VERR(ctx, "cannot create blob %s, illegal encoding beginning "
		    "with \"%s\"", vcl_name, p);
		return;
	}
	if (len == 0) {
		b->blob.len = 0;
		free(b->blob.priv);
		b->blob.priv = NULL;
		return;
	}
	b->blob.len = len;
}

VCL_BLOB __match_proto__(td_blobcode_blob_get)
vmod_blob_get(VRT_CTX, struct vmod_blobcode_blob *b)
{
	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(b, VMOD_BLOBCODE_BLOB_MAGIC);
	return &b->blob;
}

VCL_STRING
vmod_blob_encode(VRT_CTX, struct vmod_blobcode_blob *b, VCL_ENUM encs)
{
	enum encoding enc = parse_encoding(encs);

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(b, VMOD_BLOBCODE_BLOB_MAGIC);

	if (b->blob.len == 0)
		return "";

	if (b->encoding[enc] == NULL) {
		AZ(pthread_mutex_lock(&b->lock));
		if (b->encoding[enc] == NULL) {
			ssize_t len;

			switch(enc) {
			case IDENTITY:
				len = b->blob.len + 1;
				break;
			case HEXUC:
			case HEXLC:
				len = hex_encode_l(b->blob.len);
				break;
			case BASE64:
			case BASE64URL:
				len = base64_encode_l(b->blob.len);
				break;
			case BASE64URLNOPAD:
				len = base64nopad_encode_l(b->blob.len);
				break;
			default:
				WRONG("Illegal encoding");
			}
			assert(len >= 0);
			if (len == 0)
				b->encoding[enc] = "";
			else {
				b->encoding[enc] = malloc(len);
				if (b->encoding[enc] == NULL)
					ERRNOMEM(ctx, "cannot encode");
				else {
					const char *s = b->encoding[enc];
					len = encode(enc, (void *) s, len,
						     b->blob.priv, b->blob.len);
					assert(len >= 0);
					if (len == 0) {
						free((void *) s);
						b->encoding[enc] = "";
					}
					else
						*((char *)s + len) = '\0';
				}
			}
		}
		AZ(pthread_mutex_unlock(&b->lock));
	}
	return b->encoding[enc];
}

VCL_VOID __match_proto__(td_blobcode_blob__fini)
vmod_blob__fini(struct vmod_blobcode_blob **blobp)
{
	struct vmod_blobcode_blob *b;

	AN(*blobp);
	b = *blobp;
	*blobp = NULL;
	CHECK_OBJ_NOTNULL(b, VMOD_BLOBCODE_BLOB_MAGIC);
	if (b->blob.priv != NULL) {
		free(b->blob.priv);
		b->blob.priv = NULL;
	}
	for (int i = 0; i < __MAX_ENCODING; i++)
		if (b->encoding[i] != NULL && b->encoding[i][0] != '\0') {
			free((void *) b->encoding[i]);
			b->encoding[i] = NULL;
		}
	AZ(pthread_mutex_destroy(&b->lock));
	FREE_OBJ(b);
}

/* Functions */

VCL_BLOB __match_proto__(td_blobcode_decode)
vmod_decode(VRT_CTX, VCL_ENUM decs, const char *p, ...) {
	enum encoding dec = parse_encoding(decs);
	va_list ap;
	struct wb_s wb;
	struct vmod_priv *b;
	char *snap, *buf;
	ssize_t len;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	if (INIT_FINI(ctx)) {
		ILLEGAL(ctx, "decode()");
		return NULL;
	}
	CHECK_OBJ_NOTNULL(ctx->ws, WS_MAGIC);

	snap = WS_Snapshot(ctx->ws);
	if ((b = WS_Alloc(ctx->ws, sizeof(struct vmod_priv))) == NULL) {
		ERRNOMEM(ctx, "cannot decode");
		return NULL;
	}
	if (wb_create(ctx->ws, &wb) == NULL) {
		WS_Reset(ctx->ws, snap);
		ERRNOMEM(ctx, "cannot decode");
		return NULL;
	}
	buf = wb_buf(&wb);

	va_start(ap, p);
	len = decode(dec, buf, wb_space(&wb), p, ap);
	va_end(ap);

	if (len == -1) {
		err_decode(ctx, p);
		wb_reset(&wb);
		WS_Reset(ctx->ws, snap);
		return NULL;
	}
	if (len == 0) {
		wb_reset(&wb);
		WS_Reset(ctx->ws, snap);
		return null_blob;
	}
	wb_advance(&wb, len);
	WS_ReleaseP(ctx->ws, wb_buf(&wb));
	b->priv = buf;
	b->len = len;
	b->free = NULL;
	return b;
}

VCL_STRING __match_proto__(td_blobcode_encode)
vmod_encode(VRT_CTX, VCL_ENUM encs, VCL_BLOB b) {
	enum encoding enc = parse_encoding(encs);
	struct wb_s wb;
	ssize_t len;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	if (INIT_FINI(ctx)) {
		ILLEGAL(ctx, "encode()");
		return NULL;
	}
	if (b == NULL)
		return NULL;

	CHECK_OBJ_NOTNULL(ctx->ws, WS_MAGIC);
	if (wb_create(ctx->ws, &wb) == NULL) {
		ERRNOMEM(ctx, "cannot encode");
		return NULL;
	}

	len = encode(enc, wb_buf(&wb), wb_space(&wb), b->priv, b->len);

	if (len == -1) {
		ERRNOMEM(ctx, "cannot encode");
		wb_reset(&wb);
		return NULL;
	}
	if (len == 0) {
		wb_reset(&wb);
		return "";
	}
	wb_advance(&wb, len);
	return wb_finish(&wb, NULL);
}

VCL_STRING __match_proto__(td_blobcode_transcode)
vmod_transcode(VRT_CTX, VCL_ENUM decs, VCL_ENUM encs, const char *p, ...) {
	enum encoding dec = parse_encoding(decs);
	enum encoding enc = parse_encoding(encs);
	struct wb_s wb;
	ssize_t len;
	va_list ap;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	if (INIT_FINI(ctx)) {
		ILLEGAL(ctx, "transcode()");
		return NULL;
	}
	CHECK_OBJ_NOTNULL(ctx->ws, WS_MAGIC);

	/* Allocate space for the decoded blob on the stack */
	va_start(ap, p);
	size_t l = decode_l_va(dec, p, ap);
	va_end(ap);
	if (l == 0)
		return "";
	/* XXX: handle stack overflow? */
	char buf[l];

	va_start(ap, p);
	len = decode(dec, buf, l, p, ap);
	va_end(ap);

	if (len == -1) {
		err_decode(ctx, p);
		return NULL;
	}

	/*
	 * If the encoding and decoding are the same, and the decoding was
	 * legal, just return the concatenated string.
	 */
	if (enc == dec) {
		va_start(ap, p);
		VCL_STRING s = VRT_String(ctx->ws, NULL, p, ap);
		va_end(ap);
		return s;
	}

	if (wb_create(ctx->ws, &wb) == NULL) {
		ERRNOMEM(ctx, "cannot encode after decode");
		return NULL;
	}

	len = encode(enc, wb_buf(&wb), wb_space(&wb), buf, len);
	if (len == -1) {
		ERRNOMEM(ctx, "cannot encode after decode");
		wb_reset(&wb);
		return NULL;
	}
	if (len == 0) {
		wb_reset(&wb);
		return "";
	}
	wb_advance(&wb, len);
	return wb_finish(&wb, NULL);
}

VCL_STRING __match_proto__(td_blobcode_version)
vmod_version(VRT_CTX __attribute__((unused)))
{
	return VERSION;
}
